@@ -951,8 +951,21 @@ static void S_mop_class_apply_role(pTHX_ RoleEmbedding *embedding)
       ParamMeta *classparammeta;
       Newx(classparammeta, 1, struct ParamMeta);
 
-      classparammeta->field   = roleparammeta->field;
-      classparammeta->fieldix = roleparammeta->fieldix + embedding->offset;
+      classparammeta->name  = SvREFCNT_inc(roleparammeta->name);
+      classparammeta->class = roleparammeta->class;
+      classparammeta->type  = roleparammeta->type;
+
+      switch(roleparammeta->type) {
+        case PARAM_FIELD:
+          classparammeta->field.fieldmeta = roleparammeta->field.fieldmeta;
+          classparammeta->field.fieldix   = roleparammeta->field.fieldix + embedding->offset;
+          break;
+
+        case PARAM_ADJUST:
+          classparammeta->adjust.padix   = roleparammeta->adjust.padix;
+          classparammeta->adjust.defexpr = roleparammeta->adjust.defexpr; /* no refcnt on optrees */
+          break;
+      }
 
       if(klen < 0)
         hv_store_ent(dst, HeSVKEY(iter), (SV *)classparammeta, HeHASH(iter));
@@ -998,10 +1011,29 @@ static void S_apply_roles(pTHX_ ClassMeta *dstmeta, ClassMeta *srcmeta)
   }
 }
 
+void ObjectPad__check_colliding_param(pTHX_ ClassMeta *classmeta, SV *paramname)
+{
+  HV *parammap = classmeta->parammap;
+  assert(parammap);
+
+  HE *he = hv_fetch_ent(parammap, paramname, 0, 0);
+  if(!he)
+    return;
+
+  ParamMeta *colliding_parammeta = (ParamMeta *)HeVAL(he);
+  ClassMeta *origclassmeta = colliding_parammeta->class;
+
+  if(origclassmeta != classmeta)
+    croak("Already have a named constructor parameter called '%" SVf "' inherited from %" SVf,
+        SVfARG(paramname), SVfARG(origclassmeta->name));
+  else
+    croak("Already have a named constructor parameter called '%" SVf "'", SVfARG(paramname));
+}
+
 static OP *pp_alias_params(pTHX)
 {
   dSP;
-  PADOFFSET padix = PADIX_INITFIELDS_PARAMS;
+  PADOFFSET padix = PADIX_PARAMS;
 
   SV *params = POPs;
 
@@ -1081,7 +1113,7 @@ static void S_generate_initfields_method(pTHX_ ClassMeta *meta)
     op = op_append_list(OP_LIST, op,
       newPADxVOP(OP_PADSV, 0, PADIX_SELF));
     op = op_append_list(OP_LIST, op,
-      newPADxVOP(OP_PADHV, OPf_REF, PADIX_INITFIELDS_PARAMS));
+      newPADxVOP(OP_PADHV, OPf_REF, PADIX_PARAMS));
     op = op_append_list(OP_LIST, op,
       newSVOP(OP_CONST, 0, (SV *)supermeta->initfields));
 
@@ -1117,7 +1149,7 @@ static void S_generate_initfields_method(pTHX_ ClassMeta *meta)
       op = op_append_list(OP_LIST, op,
         newPADxVOP(OP_PADSV, 0, PADIX_SELF));
       op = op_append_list(OP_LIST, op,
-        newPADxVOP(OP_PADHV, OPf_REF, PADIX_INITFIELDS_PARAMS));
+        newPADxVOP(OP_PADHV, OPf_REF, PADIX_PARAMS));
       op = op_append_list(OP_LIST, op,
         newSVOP(OP_CONST, 0, (SV *)embed_cv(rolemeta->initfields, embedding)));
 
@@ -1554,11 +1586,12 @@ XS_INTERNAL(injected_constructor)
     /* Concat all the param names, in no particular order
      * TODO: consider sorting them but that's quite expensive and tricky in XS */
 
-    SV *params = newSVsv(HeSVKEY_force(he));
+    SV *params = newSVpvn("", 0);
     SAVEFREESV(params);
+    sv_catpvf(params, "'%" SVf "'", SVfARG(HeSVKEY_force(he)));
 
     while((he = hv_iternext(paramhv)))
-      sv_catpvf(params, ", %" SVf, SVfARG(HeSVKEY_force(he)));
+      sv_catpvf(params, ", '%" SVf "'", SVfARG(HeSVKEY_force(he)));
 
 #ifdef DEBUG_OVERRIDE_PLCURCOP
     PL_curcop = prevcop;
@@ -1658,6 +1691,33 @@ XS_INTERNAL(injected_DOES)
   XSRETURN_NO;
 }
 
+static OP *pp_croak_from_constructor(pTHX)
+{
+  dSP;
+
+  /* Walk up the caller stack to find the COP of the first caller; i.e. the
+   * first one that wasn't in src/class.c
+   */
+  I32 count = 0;
+  const PERL_CONTEXT *cx;
+  while((cx = caller_cx(count, NULL))) {
+    const char *copfile = CopFILE(cx->blk_oldcop);
+    if(!copfile|| strNE(copfile, "src/class.c")) {
+      PL_curcop = cx->blk_oldcop;
+      break;
+    }
+    count++;
+  }
+
+  croak_sv(POPs);
+}
+
+OP *ObjectPad__newop_croak_from_constructor(pTHX_ SV *message)
+{
+  return newUNOP_CUSTOM(&pp_croak_from_constructor, 0,
+    newSVOP(OP_CONST, 0, message));
+}
+
 ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name)
 {
   assert(type == METATYPE_CLASS || type == METATYPE_ROLE);
@@ -1732,7 +1792,7 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name)
       pad_add_name_pvs("", 0, NULL, NULL);
 
     PADOFFSET padix = pad_add_name_pvs("%params", 0, NULL, NULL);
-    if(padix != PADIX_INITFIELDS_PARAMS)
+    if(padix != PADIX_PARAMS)
       croak("ARGH: Expected that padix[%%params] = 4");
 
     intro_my();
