@@ -61,28 +61,20 @@ static void S_mop_field_set_param(pTHX_ FieldMeta *fieldmeta, SV *paramname)
   if(!classmeta->parammap)
     classmeta->parammap = newHV();
 
-  HV *parammap = classmeta->parammap;
-
-  HE *he;
-  if((he = hv_fetch_ent(parammap, paramname, 0, 0))) {
-    ParamMeta *colliding_parammeta = (ParamMeta *)HeVAL(he);
-    if(colliding_parammeta->field->class != classmeta)
-      croak("Already have a named constructor parameter called '%" SVf "' inherited from %" SVf,
-        SVfARG(paramname), SVfARG(colliding_parammeta->field->class->name));
-    else
-      croak("Already have a named constructor parameter called '%" SVf "'", SVfARG(paramname));
-  }
+  check_colliding_param(classmeta, paramname);
 
   ParamMeta *parammeta;
   Newx(parammeta, 1, struct ParamMeta);
 
   parammeta->name = SvREFCNT_inc(paramname);
-  parammeta->field = fieldmeta;
-  parammeta->fieldix = fieldmeta->fieldix;
+  parammeta->class = classmeta;
+  parammeta->type = PARAM_FIELD;
+  parammeta->field.fieldmeta = fieldmeta;
+  parammeta->field.fieldix   = fieldmeta->fieldix;
 
   fieldmeta->paramname = SvREFCNT_inc(paramname);
 
-  hv_store_ent(parammap, paramname, (SV *)parammeta, 0);
+  hv_store_ent(classmeta->parammap, paramname, (SV *)parammeta, 0);
 }
 
 SV *ObjectPad_mop_field_get_default_sv(pTHX_ FieldMeta *fieldmeta)
@@ -305,27 +297,6 @@ AV *ObjectPad_mop_field_get_attribute_values(pTHX_ FieldMeta *fieldmeta, const c
   return ret;
 }
 
-static OP *pp_croak_from_constructor(pTHX)
-{
-  dSP;
-
-  /* Walk up the caller stack to find the COP of the first caller; i.e. the
-   * first one that wasn't in src/class.c
-   */
-  I32 count = 0;
-  const PERL_CONTEXT *cx;
-  while((cx = caller_cx(count, NULL))) {
-    const char *copfile = CopFILE(cx->blk_oldcop);
-    if(!copfile|| strNE(copfile, "src/class.c")) {
-      PL_curcop = cx->blk_oldcop;
-      break;
-    }
-    count++;
-  }
-
-  croak_sv(POPs);
-}
-
 #define gen_field_init_op(fieldmeta)  S_gen_field_init_op(aTHX_ fieldmeta)
 static OP *S_gen_field_init_op(pTHX_ FieldMeta *fieldmeta)
 {
@@ -347,22 +318,21 @@ static OP *S_gen_field_init_op(pTHX_ FieldMeta *fieldmeta)
         SV *paramname = fieldmeta->paramname;
 
         if(!valueop)
-          valueop = newUNOP_CUSTOM(&pp_croak_from_constructor, 0,
-            newSVOP(OP_CONST, 0,
-              newSVpvf("Required parameter '%" SVf "' is missing for %" SVf " constructor",
-                SVfARG(paramname), SVfARG(classmeta->name))));
+          valueop = newop_croak_from_constructor(
+            newSVpvf("Required parameter '%" SVf "' is missing for %" SVf " constructor",
+              SVfARG(paramname), SVfARG(classmeta->name)));
 
         valueop = newCONDOP(0,
           /* exists $params{$paramname} */
           newUNOP(OP_EXISTS, 0,
             newBINOP(OP_HELEM, 0,
-              newPADxVOP(OP_PADHV, OPf_REF, PADIX_INITFIELDS_PARAMS),
+              newPADxVOP(OP_PADHV, OPf_REF, PADIX_PARAMS),
               newSVOP(OP_CONST, 0, SvREFCNT_inc(paramname)))),
 
           /* ? delete $params{$paramname} */
           newUNOP(OP_DELETE, 0,
             newBINOP(OP_HELEM, 0,
-              newPADxVOP(OP_PADHV, OPf_REF, PADIX_INITFIELDS_PARAMS),
+              newPADxVOP(OP_PADHV, OPf_REF, PADIX_PARAMS),
               newSVOP(OP_CONST, 0, SvREFCNT_inc(paramname)))),
 
           /* : valueop or die */
